#  Copyright (c) 2019 Markus Ressel
#  .
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#  .
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#  .
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.

import re
from typing import Pattern

from container_app_conf import ConfigEntry
from container_app_conf.util import regex_deepcopy_36_workaround

regex_deepcopy_36_workaround()


class RegexConfigEntry(ConfigEntry):
    _example = r"^[a-zA-z0-9]*$"

    def __init__(self, key_path: [str], example: any = None, description: str or None = None, default: any = None,
                 required: bool = None, secret: bool = None, flags: int or None = None):
        """
        :param flags: Regex flags to mandate or None if no flag restrictions should be made
        """
        self.flags = flags | re.UNICODE if flags is not None else None
        super().__init__(key_path, example, description, default, required, secret)

    def _value_to_type(self, value: any) -> Pattern or None:
        if value is None and self._required:
            return None

        if isinstance(value, Pattern):
            if self.flags is not None:
                if value.flags == int(self.flags):
                    return value
                else:
                    raise ValueError("Value does not match expected flags: {}".format(self.flags))
        else:
            value = str(value)
        return re.compile(value, flags=self.flags if self.flags is not None else 0)

    def _type_to_value(self, type: any) -> any:
        if type is None:
            return None
        if isinstance(type, str):
            return type
        return type.pattern



#  Copyright (c) 2019 Markus Ressel
#  .
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#  .
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#  .
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.


class ConfigFormatter:
    """
    Allows config entries to be formatted into a string
    """

    def format(self, data: dict) -> str:
        """
        Formats the given entry data
        :param data: entries to format
        :return: formatted string
        """
        raise NotImplementedError()


class SimpleFormatter(ConfigFormatter):
    """
    Prints all config entries in a human readable manner
    """

    def format(self, data: dict) -> str:
        return "\n".join(self._format(data)).strip()

    def _format(self, data: dict, prefix: str = "") -> [str]:
        """
        Recursively formats the dictionary
        :param data:
        :return:
        """
        lines = []
        for key, value in data.items():
            output = prefix + key
            if isinstance(value, dict):
                lines.extend(self._format(value, "{}->".format(output)))
            else:
                lines.append(output + ": {}".format(value))
        return lines



#  Copyright (c) 2019 Markus Ressel
#  .
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#  .
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#  .
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
from py_range_parse import Range

from container_app_conf import ConfigEntry


class FloatConfigEntry(ConfigEntry):
    _example = "3.1415926535"

    def __init__(self, key_path: [str], example: any = None, description: str or None = None, default: any = None,
                 required: bool = None, secret: bool = None, range: Range = None):
        self.range = range
        super().__init__(key_path, example, description, default, required, secret)

    def _value_to_type(self, value: any) -> float or None:
        """
        Tries to permissively convert the given value to a float.
        :param value: the value to parse
        :return: the parsed float value
        """
        if isinstance(value, str) and '%' == value[-1]:
            parsed_value = float(value[:-1]) / 100.0
        else:
            parsed_value = float(value)

        if self.range is not None and parsed_value not in self.range:
            raise ValueError("Value not in range: {} not in {}".format(parsed_value, self.range))
        return parsed_value

    def _type_to_value(self, type: any) -> any:
        if type is None:
            return None
        return float(type)



#  Copyright (c) 2019 Markus Ressel
#  .
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#  .
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#  .
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
import copy
import os

from container_app_conf import ConfigEntry
from container_app_conf.source import DataSource


class EnvSource(DataSource):
    """
    Data source utilizing environment variables
    """
    KEY_SPLIT_CHAR = "_"

    def has(self, entry: ConfigEntry) -> bool:
        original_key = self.env_key(entry)
        normalized_key = original_key.replace('-', '_')
        return original_key in self.root.keys() or normalized_key in self.root.keys()

    def get(self, entry: ConfigEntry) -> str or None:
        original_key = self.env_key(entry)
        normalized_key = original_key.replace('-', '_')
        return self.root.get(original_key, self.root.get(normalized_key, None))

    @staticmethod
    def env_key(entry: ConfigEntry) -> str:
        return EnvSource.KEY_SPLIT_CHAR.join(entry.key_path).upper()

    def _load(self) -> dict:
        return copy.deepcopy(os.environ)



#  Copyright (c) 2019 Markus Ressel
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#

from setuptools import setup, find_packages

VERSION_NUMBER = "5.2.2"

DEVELOPMENT_STATUS = "Development Status :: 5 - Production/Stable"
VERSION_NAME = VERSION_NUMBER


def readme_type() -> str:
    import os
    if os.path.exists("README.rst"):
        return "text/x-rst"
    if os.path.exists("README.md"):
        return "text/markdown"


def readme() -> [str]:
    if readme_type() == "text/markdown":
        file_name = "README.md"
    else:
        file_name = "README.rst"

    with open(file_name) as f:
        return f.read()


def locked_requirements(section):
    """
    Look through the 'Pipfile.lock' to fetch requirements by section.
    """
    import json

    with open('Pipfile.lock') as pip_file:
        pipfile_json = json.load(pip_file)

    if section not in pipfile_json:
        print("{0} section missing from Pipfile.lock".format(section))
        return []

    return [package
            for package, detail in pipfile_json[section].items()]


setup(
    name='container_app_conf',
    version=VERSION_NAME,
    description='Convenient configuration of containerized applications',
    long_description=readme(),
    long_description_content_type=readme_type(),
    license='MIT',
    author='Markus Ressel',
    author_email='mail@markusressel.de',
    url='https://github.com/markusressel/container-app-conf',
    packages=find_packages(),
    classifiers=[
        DEVELOPMENT_STATUS,
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7'
    ],
    install_requires=locked_requirements('default'),
    tests_require=locked_requirements('develop'),
)
